import unittest

from tea.dsa.singleton import Singleton


class TestMultiConfig(unittest.TestCase):
    def test_subclassing(self):
        class MySingleton(Singleton):
            def __init__(self, data):
                self.data = data

        first = MySingleton("First")
        second = MySingleton("Second")
        self.assertEqual(first.data, "First")
        self.assertEqual(second.data, "First")
        self.assertIs(first, second)



__all__ = [
    "Process",
    "ExecutableNotFound",
    "kill",
    "find",
    "get_processes",
    "execute",
    "execute_no_demux",
    "execute_and_report",
]

from tea.process.process import Process, ExecutableNotFound, kill
from tea.process.wrappers import (
    find,
    get_processes,
    execute,
    execute_no_demux,
    execute_and_report,
)



import pytest

from tea.version import Version


@pytest.fixture
def version():
    return Version(3, 4, 5)


def test_version_fields(version):
    assert version.major == 3
    assert version.minor == 4
    assert version.build == 5


def test_version_string(version):
    assert str(version) == "3.4.5"



import os
import re
import sys
import time
import pytest
from tea.process import (
    Process,
    ExecutableNotFound,
    execute,
    execute_no_demux,
    execute_and_report as er,
)


WRITER = """
import sys
import time

sys.{out}.write("foo")
sys.{out}.flush()
time.sleep(1)
sys.{out}.write("bar")
"""

WRITE_BOTH = """
import sys

sys.stdout.write("foo")
sys.stdout.flush()
sys.stderr.write("bar")
sys.stderr.flush()
"""

PRINT_VAR = """
import os

print(os.environ.get("{var}", ""))
"""


@pytest.fixture
def process():
    return Process(["sleep", "1"])


def test_start(process):
    assert not process.is_running
    process.start()
    assert process.is_running
    process.wait()
    assert not process.is_running


def test_kill(process):
    assert not process.is_running
    start = time.time()
    process.start()
    assert process.is_running
    process.kill()
    assert not process.is_running
    end = time.time()
    assert end - start < 2


def test_write():
    p = Process(
        [sys.executable, "-c", """a = input(); print('Said: ' + a)"""],
    )
    p.start()
    p.write("my hello text")
    p.wait()
    assert re.match(r"^Said: my hello text\s*$", p.read())
    assert re.match(r"^$", p.eread())


def test_read():
    p = Process([sys.executable, "-c", WRITER.format(out="stdout")])
    p.start()
    time.sleep(0.5)
    assert p.read() == "foo"
    assert p.eread() == ""
    p.wait()
    assert p.read() == "bar"
    assert p.eread() == ""


def test_eread():
    p = Process([sys.executable, "-c", WRITER.format(out="stderr")])
    p.start()
    time.sleep(0.5)
    assert p.read() == ""
    assert p.eread() == "foo"
    p.wait()
    assert p.read() == ""
    assert p.eread() == "bar"


def test_stdout(tmpdir):
    stdout = tmpdir.mkdir("__tests__").join("std.out")
    p = Process(
        [sys.executable, "-c", WRITER.format(out="stdout")],
        stdout=stdout.strpath,
    )
    p.start()
    p.wait()
    assert p.read() == "foobar"
    assert stdout.open("r", encoding="utf-8").read() == "foobar"


def test_stderr(tmpdir):
    stderr = tmpdir.mkdir("__tests__").join("std.err")
    p = Process(
        [sys.executable, "-c", WRITER.format(out="stderr")],
        stderr=stderr.strpath,
    )
    p.start()
    p.wait()
    assert p.eread() == "foobar"
    assert stderr.open("r", encoding="utf-8").read() == "foobar"


def test_environment():
    p = Process(
        [sys.executable, "-c", PRINT_VAR.format(var="MY_VAR")],
        env={"MY_VAR": "My value"},
    )
    p.start()
    p.wait()
    assert p.exit_code == 0
    assert re.match(r"^My value\s*$", p.read())
    assert re.match(r"^$", p.eread())


def test_override_environment():
    status, output, error = execute(
        [sys.executable, "-c", PRINT_VAR.format(var="PATH")],
        env={"PATH": "PATH"},
    )
    assert status == 0
    assert re.match(r"^PATH\s*$", output)
    assert re.match(r"^$", error)
    os.environ["FOO"] = "foo"
    status, output, error = execute(
        [sys.executable, "-c", PRINT_VAR.format(var="FOO")]
    )
    assert status == 0
    assert re.match(r"^foo\s*$", output)
    assert re.match(r"^$", error)
    status, output, error = execute(
        [sys.executable, "-c", PRINT_VAR.format(var="FOO")], env={"FOO": "bar"}
    )
    assert status == 0
    assert re.match(r"^bar\s*$", output)
    assert re.match(r"^$", error)


def test_working_dir():
    working_dir = sys.exec_prefix.strip(os.pathsep)
    p = Process(
        [sys.executable, "-c", """import os; print(os.getcwd())"""],
        working_dir=working_dir,
    )
    p.start()
    p.wait()
    assert p.read().strip() == working_dir


def test_execute_with_error():
    status, output, error = execute(
        [sys.executable, "-c", "import sys; sys.exit(2)"]
    )
    assert status == 2
    assert re.match(r"^$", output)
    assert re.match(r"^$", error)


def test_execute_with_success():
    status, output, error = execute(["echo", "Hello world"])
    assert status == 0
    assert re.match(r"^Hello world\s*$", output)
    assert re.match(r"^$", error)


def test_execute_demux():
    status, output, error = execute([sys.executable, "-c", WRITE_BOTH])
    assert status == 0
    assert output == "foo"
    assert error == "bar"


def test_execute_no_demux():
    status, output = execute_no_demux([sys.executable, "-c", WRITE_BOTH])
    assert status == 0
    assert output == "foobar"


def test_execute_and_report():
    assert er([sys.executable, "-c", "import sys; sys.exit(0)"])
    assert not er([sys.executable, "-c", "import sys; sys.exit(1)"])
    assert not er([sys.executable, "-c", "import sys; sys.exit(-1)"])
    assert not er([sys.executable, "-c", "import sys; sys.exit(2)"])


def test_not_existing_command():
    pytest.raises(ExecutableNotFound, lambda: execute("non_existing_command"))



import io
import os
import time
import posix
import signal
import logging
import threading
import subprocess
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Optional, Dict, Union, List

from tea.errors import TeaError


logger = logging.getLogger(__name__)


class ExecutableNotFound(TeaError):
    def __init__(self, command):
        self.command = command
        super().__init__(message=f"Executable not found: {command}")


def _create_env(env):
    full_env = {str(key): str(value) for key, value in os.environ.items()}
    if env is not None:
        full_env.update({str(key): str(value) for key, value in env.items()})
    return full_env


def kill(pid):
    """Kills a process by it's process ID.

    Args:
        pid (int): Process ID of the process to kill.
    """
    if pid == posix.getpgid(pid):
        os.killpg(pid, signal.SIGKILL)
    else:
        os.kill(pid, signal.SIGKILL)


class Process:
    r"""Process class.

    Simple example of Process class usage can be::

        >>> from tea.process import Process
        >>> p = Process('python', ['-c', 'import time;time.sleep(5);print(3)'])
        >>> p.start()
        >>> p.is_running
        True
        >>> p.wait()
        True
        >>> p.read()
        '3\\n'
        >>> p.eread()
        ''
    """

    def __init__(
        self,
        command: Union[str, List[str]],
        env: Optional[Dict[str, str]] = None,
        stdout: Optional[Union[str, Path]] = None,
        stderr: Optional[Union[str, Path]] = None,
        demux: bool = True,
        redirect_output: bool = True,
        working_dir: Optional[str] = None,
        encoding: str = "utf-8",
    ):
        """Create a Process object.

        The only required parameter is the command to execute. It's important
        to note that the constructor only initializes the class, it doesn't
        executes the process. To actually execute the process you have to call
        the `start` method.

        Args:
            command: Path to the executable file or a list with the full
                command and it's arguments.
            env: Optional additional environment variables that will be added
                to the subprocess environment or that override currently set
                environment variables.
            stdout: Path to the file to which standard output would be
                redirected.
            stderr: Path to the file to which standard error would be
                redirected.
            demux: Demux stdout and stderr. Default: `True`. If demux is set to
                `False` `eread` will always return an empty string and all
                stderr will be redirected to stdout.
            redirect_output: `True` if you want to be able to get the standard
                output and the standard error of the subprocess, otherwise it
                will be redirected to /dev/null. If stdout or stderr are
                provided redirect_output will automatically be set to `True`.
            working_dir: Set the working directory from which the process will
                be started.
            encoding: Encoding to use to decode the standard output and
                standard error.
        """
        self._commandline = [command] if isinstance(command, str) else command
        self._env = env
        self._process = None
        self._wait_thread = None
        self._pid = None
        self._immutable = False
        self._working_dir = working_dir
        # Demux, redirect and ecnoding
        self._demux = demux
        self._redirect_output = stdout or stderr or redirect_output
        self._encoding = encoding
        # stdin
        self._stdin = None
        # stdout
        self._stdout_tmp = None
        self._stdout = Path(stdout).absolute() if stdout else None
        self._stdout_reader = None
        self._stdout_writer = None
        # stderr
        self._stderr_tmp = None
        self._stderr = Path(stderr).absolute() if stderr else None
        self._stderr_reader = None
        self._stderr_writer = None

    def __open_files(self):
        if self._redirect_output:
            # stdin
            self._stdin = subprocess.PIPE
            # stdout
            if self._stdout:
                self._stdout_writer = io.open(self._stdout, "wb")
                self._stdout_reader = io.open(self._stdout, "rb")
            else:
                self._stdout_tmp = NamedTemporaryFile(mode="wb")
                self._stdout_writer = self._stdout_tmp.file
                self._stdout_reader = io.open(self._stdout_tmp.name, "rb")
            # stderr
            if self._demux:
                if self._stderr:
                    self._stderr_writer = io.open(self._stderr, "wb")
                    self._stderr_reader = io.open(self._stderr, "rb")
                else:
                    self._stderr_tmp = NamedTemporaryFile(mode="wb")
                    self._stderr_writer = self._stderr_tmp.file
                    self._stderr_reader = io.open(self._stderr_tmp.name, "rb")
            else:
                self._stderr_writer = subprocess.STDOUT
        else:
            self._stdout_writer = io.open(os.devnull, "wb")
            self._stderr_writer = subprocess.STDOUT

    def __is_open(self, f):
        return f is not None and hasattr(f, "closed") and not f.closed

    def __close_write_files(self):
        # Close stdout
        if self.__is_open(self._stdout_tmp):
            self._stdout_tmp.close()
        if self.__is_open(self._stderr_writer):
            self._stdout_writer.close()
        # Close stderr
        if self.__is_open(self._stderr_tmp):
            self._stderr_tmp.close()
        if self._demux and self.__is_open(self._stderr_writer):
            self._stderr_writer.close()

    def __close_files(self):
        # Close read files
        if self.__is_open(self._stdout_reader):
            self._stdout_reader.close()
        if self._demux and self.__is_open(self._stderr_reader):
            self._stderr_reader.close()
        # Close write files
        self.__close_write_files()

    def __process_wait(self):
        self._process.wait()
        self.__close_write_files()

    @classmethod
    def immutable(cls, pid, command):
        """Create an immutable process object used for listing processes."""
        p = cls(command[0], command[1:])
        p._pid = pid
        p._immutable = True
        return p

    @property
    def command(self):
        """Command."""
        return self._commandline[0]

    @property
    def arguments(self):
        """Arguments."""
        return self._commandline[1:]

    @property
    def command_line(self):
        """Full command line."""
        return self._commandline

    def start(self):
        """Start the process."""
        if self._immutable:
            raise NotImplementedError

        # Open all redirects
        self.__open_files()

        try:
            self._process = subprocess.Popen(
                self._commandline,
                stdin=self._stdin,
                stdout=self._stdout_writer,
                stderr=self._stderr_writer,
                env=_create_env(self._env),
                cwd=self._working_dir,
            )
        except OSError:
            raise ExecutableNotFound(command=self.command)
        self._wait_thread = threading.Thread(
            target=self.__process_wait, daemon=True
        )
        self._wait_thread.start()

    def kill(self):
        """Kill the process if it's running."""
        try:
            if self._process is not None:
                kill(self.pid)
                self._wait_thread.join()
                self._process = None
                return True
            elif self._immutable:
                kill(self.pid)
                return True
            else:
                return None
        except OSError:
            return False

    def wait(self, timeout: Optional[int] = None):
        """Wait for the process to finish.

        It will wait for the process to finish running. If the timeout is
        provided, the function will wait only `timeout` amount of seconds and
        then return to it's caller.

        Args:
            timeout: `None` if you want to wait to wait until the process
                actually finishes, otherwise it will wait just the `timeout`
                number of seconds.

        Returns:
            bool: Return value only makes sense if you provided the timeout
                parameter. It will indicate if the process actually finished in
                the amount of time specified, i.e. if the we specify 3 seconds
                and the process actually stopped after 3 seconds it will return
                `True` otherwise it will return `False`.
        """
        if self._immutable:
            raise NotImplementedError

        if timeout is not None:
            current_time = time.time()
            while time.time() - current_time < (timeout * 1000):
                if not self._process.is_running:
                    return True
                time.sleep(0.1)
            return False
        else:
            while self.is_running:
                time.sleep(0.1)
            return True

    @property
    def is_running(self):
        """Indicate if the process is still running.

        Returns:
            bool: `True` if the process is still running `False` otherwise.
        """
        if self._immutable:
            raise NotImplementedError

        if self._process is None or self._process.returncode is not None:
            return False
        return True

    @property
    def pid(self):
        """PID of the process if it is running.

        Returns:
            int: Process id of the running process.
        """
        return self._pid if self._immutable else self._process.pid

    @property
    def exit_code(self) -> Optional[int]:
        """Exit code if the process has finished running.

        Returns:
            Optional[int]: Exit code or `None` if the process is still running.
        """
        if self._immutable:
            raise NotImplementedError

        if self.is_running:
            return None
        return self._process.returncode

    def write(self, string: str):
        """Write a string to the process standard input.

        Args:
            string: String to write to the process standard input.
        """
        if self._immutable:
            raise NotImplementedError

        if self._redirect_output:
            if string[-1] != "\n":
                string += "\n"
            self._process.stdin.write(string.encode(self._encoding))
            self._process.stdin.flush()

    def read(self) -> str:
        """Read from the process standard output.

        Returns:
            str: The data process has written to the standard output if it has
                written anything. If it hasn't or you already read all the data
                process wrote, it will return an empty string.
        """
        if self._immutable:
            raise NotImplementedError

        if self._redirect_output:
            return self._stdout_reader.read().decode(self._encoding)
        return ""

    def eread(self) -> str:
        """Read from the process standard error.

        Returns:
            str: The data process has written to the standard error if it has
                written anything. If it hasn't or you already read all the data
                process wrote, it will return an empty string.
        """
        if self._immutable:
            raise NotImplementedError

        if self._redirect_output and self._demux:
            return self._stderr_reader.read().decode("utf-8")
        return ""

    def __del__(self):
        self.__close_files()

    def __str__(self):
        return f"Process(pid={self.pid}, command={self.command})"

    __repr__ = __str__



import os
from unittest import mock

import pytest

from tea.dsa.config import Config


DATA_DIR = os.path.join(os.path.dirname(__file__), "data")

DICT_DATA = {"foo": {"bar": {"baz": 1}, "baz": 2}, "bar": {"baz": 3}, "baz": 4}

INI_DATA = open(os.path.join(DATA_DIR, "config.ini")).read()
JSON_DATA = open(os.path.join(DATA_DIR, "config.json")).read()


@pytest.fixture
def config():
    return Config(data=DICT_DATA)


def is_subset(dict1, dict2):
    """Is dict1 subset of dict2."""
    for key, value in dict1.items():
        if key not in dict2 or value != dict2[key]:
            return False
    return True


def check_values(config, foo=True, bar=True):
    if foo:
        assert is_subset({"bar": {"baz": 1}, "baz": 2}, config.get("foo"))
        assert is_subset({"baz": 1}, config.get("foo.bar"))
        assert config.get("foo.bar.baz") == 1
        assert config.get("foo.baz") == 2
    if bar:
        assert is_subset({"baz": 3}, config.get("bar"))
        assert config.get("bar.baz") == 3
    assert config.get("baz") == 4


def test_data_dict():
    check_values(Config(data=DICT_DATA))


def test_data_json():
    check_values(Config(data=JSON_DATA, fmt=Config.JSON))


def test_file_json():
    check_values(Config(filename=os.path.join(DATA_DIR, "config.json")))


def test_non_existing_file():
    config = Config(filename=os.path.join(DATA_DIR, "non_existin.json"))
    assert config.data == {}


def test_unsupported_format():
    config = Config(filename=os.path.join(DATA_DIR, "config.ini"))
    assert config.data == {}
    config = Config(data=INI_DATA, fmt="INI")
    assert config.data == {}


def test_invalid_format():
    config = Config(
        filename=os.path.join(DATA_DIR, "config.ini"), fmt=Config.JSON
    )
    assert config.data == {}
    config = Config(data=INI_DATA, fmt=Config.JSON)
    assert config.data == {}


@mock.patch("os.path.isdir", return_value=False)
@mock.patch("os.makedirs")
@mock.patch("io.open")
def test_save_json(io_open, makedirs, isdir):
    filename = "some_filename"
    dirname = os.path.abspath(os.path.dirname(filename))
    c = Config(data=JSON_DATA, fmt=Config.JSON)
    c.filename = filename
    c.save()
    isdir.assert_called_with(dirname)
    makedirs.assert_called_with(dirname, 0o755)
    io_open.assert_called_with(filename, "w", encoding="utf-8")
    io_open.result.write.asswert_called_with(JSON_DATA)


@mock.patch("os.path.isdir")
@mock.patch("os.makedirs")
@mock.patch("io.open")
def test_save_unsupported(io_open, makedirs, isdir):
    isdir.return_value = True
    filename = "some_filename"
    dirname = os.path.abspath(os.path.dirname(filename))
    c = Config(data=JSON_DATA, fmt=Config.JSON)
    c.filename = filename
    c.fmt = "INI"
    c.save()
    isdir.assert_called_with(dirname)
    assert not makedirs.called
    assert not io_open.called


def test_add_first_level(config):
    config.set("first", 5)
    assert config.get("first") == 5
    # the rest should be untouched
    check_values(config)


def test_add_second_level(config):
    config.set("first.second", 6)
    assert config.get("first") == {"second": 6}
    assert config.get("first.second") == 6
    # the rest should be untouched
    check_values(config)


def test_add_in_existing_first_level_one_level(config):
    config.set("foo.first", 7)
    assert config.get("foo.first") == 7
    # the rest should be untouched
    check_values(config)


def test_add_in_existing_first_level_two_levels(config):
    config.set("foo.first.second", 8)
    assert config.get("foo.first") == {"second": 8}
    assert config.get("foo.first.second") == 8
    # the rest should be untouched
    check_values(config)


def test_add_in_existing_second_level_one_level(config):
    config.set("foo.bar.first", 9)
    assert config.get("foo") == {"bar": {"baz": 1, "first": 9}, "baz": 2}
    assert config.get("foo.bar") == {"baz": 1, "first": 9}
    assert config.get("foo.bar.first") == 9
    # the rest should be untouched
    check_values(config, foo=False)


def test_add_in_existing_second_level_two_levels(config):
    config.set("foo.bar.first.second", 10)
    assert config.get("foo") == {
        "bar": {"baz": 1, "first": {"second": 10}},
        "baz": 2,
    }
    assert config.get("foo.bar") == {"baz": 1, "first": {"second": 10}}
    assert config.get("foo.bar.first") == {"second": 10}
    assert config.get("foo.bar.first.second") == 10
    # the rest should be untouched
    check_values(config, foo=False)


def test_list_insertion(config):
    config.set("foo.l", [])
    assert config.get("foo.l") == []
    config.insert("foo.l", 1)
    assert config.get("foo.l") == [1]
    assert config.get("foo.l.0") == 1
    config.insert("foo.l", 0, 0)
    assert config.get("foo.l") == [0, 1]
    assert config.get("foo.l.0") == 0
    assert config.get("foo.l.1") == 1
    config.set("foo.l.0", 2)
    assert config.get("foo.l") == [2, 1]
    assert config.get("foo.l.0") == 2
    assert config.get("foo.l.1") == 1


def test_insertion_in_non_list(config):
    pytest.raises(KeyError, lambda: config.insert("foo", 0))


def test_set_first_level(config):
    config.set("foo", 11)
    assert config.get("foo") == 11
    # the rest should be untouched
    check_values(config, foo=False)


def test_set_second_level(config):
    config.set("foo.bar", 12)
    assert config.get("foo") == {"bar": 12, "baz": 2}
    assert config.get("foo.bar") == 12
    # the rest should be untouched
    check_values(config, foo=False)


def test_set_third_level(config):
    config.set("foo.bar.baz", 13)
    assert config.get("foo") == {"bar": {"baz": 13}, "baz": 2}
    assert config.get("foo.bar") == {"baz": 13}
    assert config.get("foo.bar.baz") == 13
    # the rest should be untouched
    check_values(config, foo=False)


def test_delete_first_level(config):
    config.delete("foo")
    assert config.get("foo") is None
    # the rest should be untouched
    check_values(config, foo=False)


def test_delete_second_level(config):
    config.delete("foo.bar")
    assert config.get("foo") == {"baz": 2}
    assert config.get("foo.bar") is None
    # the rest should be untouched
    check_values(config, foo=False)


def test_delete_third_level(config):
    config.delete("foo.bar.baz")
    assert config.get("foo") == {"bar": {}, "baz": 2}
    assert config.get("foo.bar") == {}
    assert config.get("foo.bar.baz") is None
    # the rest should be untouched
    check_values(config, foo=False)


def test_delete_from_list(config):
    config.set("foo.l", [1, 2, 3])
    config.delete("foo.l.1")
    assert config.get("foo.l") == [1, 3]
    del config["foo.l.0"]
    assert config.get("foo.l") == [3]


def test_safe_delete(config):
    config.delete("non_existent")
    check_values(config)


def test_unsafe_key_error(config):
    pytest.raises(KeyError, lambda: config["nonexistant"])
    pytest.raises(KeyError, lambda: config["non.existant"])
    pytest.raises(KeyError, lambda: config["foo.nonexistant"])
    pytest.raises(KeyError, lambda: config["foo.non.existant"])
    pytest.raises(KeyError, lambda: config["foo.bar.nonexistant"])
    pytest.raises(KeyError, lambda: config["foo.bar.non.existant"])
    pytest.raises(KeyError, lambda: config["foo.bar.baz.nonexistant"])
    pytest.raises(KeyError, lambda: config["foo.bar.baz.non.existant"])


def test_unsafe_index_error(config):
    config["list"] = [1, 2]
    assert config["list.0"] == 1
    assert config["list.1"] == 2
    assert config["list.-1"] == 2
    assert config["list.-2"] == 1
    pytest.raises(IndexError, lambda: config["list.2"])
    pytest.raises(IndexError, lambda: config["list.-3"])


def test_keys(config):
    assert config.keys() == set(["foo", "bar", "baz"])


def test_contains(config):
    # Existing
    assert "foo" in config
    assert "foo.bar" in config
    assert "foo.bar.baz" in config
    assert "foo.baz" in config
    assert "bar" in config
    assert "bar.baz" in config
    assert "baz" in config
    # Non existing
    assert "oof" not in config
    assert "bar.foo" not in config
    assert "baz.foo" not in config
    assert "baz.bar" not in config
    assert "foo.baz.bar" not in config



"""Context library - providing usefull context managers."""

import contextlib


@contextlib.contextmanager
def suppress(*exceptions):
    """Ignore an exception or exception list.

    Usage::

        with suppress(OSError):
            os.remove('filename.txt')
    """
    try:
        yield
    except exceptions:
        pass



import sys
import types
import logging
import pkgutil
import importlib
import traceback


logger = logging.getLogger(__name__)


def _package(module):
    """Hack for python2 since it does not have __package__ always set up.

    Args:
        module: Module for which we want the package name
    Returns:
        str: Model's package name
    """
    return (
        module.__name__ if module.__package__ is None else module.__package__
    )


def get_object(path="", obj=None):
    """Return an object from a dot path.

    Path can either be a full path, in which case the `get_object` function
    will try to import the modules in the path and follow it to the final
    object. Or it can be a path relative to the object passed in as the second
    argument.

    Args:
        path (str): Full or relative dot path to the desired object
        obj (object): Starting object. Dot path is calculated relatively to
            this object.

    Returns:
        Object at the end of the path, or list of non hidden objects if we use
        the star query.

    Example for full paths::

        >>> get_object('os.path.join')
        <function join at 0x1002d9ed8>
        >>> get_object('tea.process')
        <module 'tea.process' from 'tea/process/__init__.pyc'>

    Example for relative paths when an object is passed in::

        >>> import os
        >>> get_object('path.join', os)
        <function join at 0x1002d9ed8>

    Example for a star query. (Star query can be used only as the last element
    of the path::

        >>> get_object('tea.dsa.*')
        []
        >>> get_object('tea.dsa.singleton.*')
        [<class 'tea.dsa.singleton.Singleton'>,
         <class 'tea.dsa.singleton.SingletonMetaclass'>
         <module 'six' from '...'>]
        >>> get_object('tea.dsa.*')
        [<module 'tea.dsa.singleton' from '...'>]    # Since we imported it
    """
    if not path:
        return obj
    path = path.split(".")
    if obj is None:
        obj = importlib.import_module(path[0])
        path = path[1:]
    for item in path:
        if item == "*":
            # This is the star query, returns non hidden objects
            return [
                getattr(obj, name)
                for name in dir(obj)
                if not name.startswith("__")
            ]
        if isinstance(obj, types.ModuleType):
            submodule = "{}.{}".format(_package(obj), item)
            try:
                obj = importlib.import_module(submodule)
            except Exception as import_error:
                try:
                    obj = getattr(obj, item)
                except Exception:
                    # FIXME: I know I should probably merge the errors, but
                    #        it's easier just to throw the import error since
                    #        it's most probably the one user wants to see.
                    #        Create a new LoadingError and throw a combination
                    #        of the import error and attribute error.
                    raise import_error
        else:
            obj = getattr(obj, item)
    return obj


class Loader(object):
    """Module loader class loads recursively a module and all it's submodules.

    Loaded modules will be stored in the ``modules`` attribute of the loader as
    a dictionary of {module_path: module} key, value pairs.

    Errors accounted during the loading process will not stop the loading
    process. They will be stored in the ``errors`` attribute of the loader as a
    dictionary of {module_path: exception} key, value pairs.

    Usage::

        loader = Loader()
        loader.load('foo')
        loader.load('baz.bar', 'boo')

        import baz
        loader.load(baz)
    """

    def __init__(self):
        self.modules = {}
        self.errors = {}

    def load(self, *modules):
        """Load one or more modules.

        Args:
            modules: Either a string full path to a module or an actual module
                object.
        """
        for module in modules:
            if isinstance(module, str):
                try:
                    module = get_object(module)
                except Exception as e:
                    self.errors[module] = e
                    continue
            self.modules[module.__package__] = module
            for (loader, module_name, is_pkg) in pkgutil.walk_packages(
                module.__path__
            ):
                full_name = "{}.{}".format(_package(module), module_name)
                try:
                    self.modules[full_name] = get_object(full_name)
                    if is_pkg:
                        self.load(self.modules[full_name])
                except Exception as e:
                    self.errors[full_name] = e


def load_subclasses(klass, modules=None):
    """Load recursively all all subclasses from a module.

    Args:
        klass (str or list of str): Class whose subclasses we want to load.
        modules: List of additional modules or module names that should be
            recursively imported in order to find all the subclasses of the
            desired class. Default: None

    FIXME: This function is kept only for backward compatibility reasons, it
        should not be used. Deprecation warning should be raised and it should
        be replaces by the ``Loader`` class.
    """
    if modules:
        if isinstance(modules, str):
            modules = [modules]
        loader = Loader()
        loader.load(*modules)
    return klass.__subclasses__()


def get_exception():
    """Return full formatted traceback as a string."""
    trace = ""
    exception = ""
    exc_list = traceback.format_exception_only(
        sys.exc_info()[0], sys.exc_info()[1]
    )
    for entry in exc_list:
        exception += entry
    tb_list = traceback.format_tb(sys.exc_info()[2])
    for entry in tb_list:
        trace += entry
    return "%s\n%s" % (exception, trace)


def cmp(x, y):
    """Compare function from python2."""
    return (x > y) - (x < y)



import unittest
from unittest import mock

from tea import shell


class TestGoto(unittest.TestCase):
    @mock.patch("tea.shell.os")
    def test_existing(self, os):
        os.getcwd.return_value = "foo"
        os.path.abspath.side_effect = lambda path: path
        os.path.isdir.return_value = True
        with shell.goto("some_directory") as ok:
            self.assertTrue(ok)
            os.getcwd.assert_called_with()
            os.path.abspath.assert_called_with("some_directory")
            os.path.isdir.assert_called_with("some_directory")
            os.chdir.assert_called_with("some_directory")
        os.chdir.assert_called_with("foo")

    @mock.patch("tea.shell.os")
    def test_non_existing_no_create(self, os):
        os.getcwd.return_value = "foo"
        os.path.abspath.side_effect = lambda path: path
        os.path.isdir.return_value = False
        with shell.goto("some_directory") as ok:
            self.assertFalse(ok)
            os.getcwd.assert_called_with()
            os.path.abspath.assert_called_with("some_directory")
            os.path.isdir.assert_called_with("some_directory")
            assert not os.chdir.called
        assert not os.chdir.called

    @mock.patch("tea.shell.os")
    def test_non_existing_create(self, os):
        os.getcwd.return_value = "foo"
        os.path.abspath.side_effect = lambda path: path
        os.path.isdir.return_value = False
        with shell.goto("some_directory", True) as ok:
            self.assertTrue(ok)
            os.getcwd.assert_called_with()
            os.path.abspath.assert_called_with("some_directory")
            os.path.isdir.assert_called_with("some_directory")
            os.makedirs.assert_called_with("some_directory", 0o755)
            os.chdir.assert_called_with("some_directory")
        os.chdir.assert_called_with("foo")

    @mock.patch("tea.shell.os")
    def test_non_existing_create_failed(self, os):
        os.getcwd.return_value = "foo"
        os.path.abspath.side_effect = lambda path: path
        os.path.isdir.return_value = False
        os.makedirs.side_effect = Exception("Failed to create")
        with shell.goto("some_directory", True) as ok:
            self.assertFalse(ok)
            os.getcwd.assert_called_with()
            os.path.abspath.assert_called_with("some_directory")
            os.path.isdir.assert_called_with("some_directory")
            os.makedirs.assert_called_with("some_directory", 0o755)
            assert not os.chdir.called
        assert not os.chdir.called

    @mock.patch("tea.shell.os")
    def test_with_throws_an_exception(self, os):
        os.getcwd.return_value = "foo"
        os.path.abspath.side_effect = lambda path: path
        os.path.isdir.return_value = True
        with self.assertRaises(Exception):
            with shell.goto("some_directory") as ok:
                self.assertTrue(ok)
                os.getcwd.assert_called_with()
                os.path.abspath.assert_called_with("some_directory")
                os.path.isdir.assert_called_with("some_directory")
                os.chdir.assert_called_with("some_directory")
                raise Exception("Failed")
        os.chdir.assert_called_with("foo")



import unittest
from tea.dsa.config import MultiConfig, Config


class TestMultiConfig(unittest.TestCase):
    dict_first = {"foo": {"bar": {"baz": 1}, "baz": 2}, "bar": 3, "baz": 4}
    dict_second = {
        "foo": {"bar": {"deep": 5}, "baz": 6, "test": 7},
        "bar": {"baz": 8},
        "first": 9,
    }

    json_first = '{"foo": {"bar": {"baz": 1}, "baz": 2}, "bar": 3, "baz": 4}'
    json_second = """{"foo": {"bar": {"deep": 5}, "baz": 6, "test": 7},
                       "bar": {"baz": 8}, "first": 9}"""

    def check_structure(self, c):
        self.assertEqual(c.get("foo.bar.baz"), 1)
        self.assertEqual(c.get("foo.bar.deep"), 5)
        self.assertEqual(c.get("foo.baz"), 6)
        self.assertEqual(c.get("foo.test"), 7)
        self.assertEqual(c.get("bar.baz"), 8)
        self.assertEqual(c.get("baz"), 4)
        self.assertEqual(c.get("first"), 9)

    def test_dict_dict(self):
        c = MultiConfig(data=self.dict_first)
        c.attach(data=self.dict_second)
        self.check_structure(c)

    def test_dict_json(self):
        c = MultiConfig(data=self.dict_first)
        c.attach(data=self.json_second, fmt=Config.JSON)
        self.check_structure(c)

    def test_json_dict(self):
        c = MultiConfig(data=self.json_first, fmt=Config.JSON)
        c.attach(data=self.dict_second)
        self.check_structure(c)

    def test_json_json(self):
        c = MultiConfig(data=self.json_first, fmt=Config.JSON)
        c.attach(data=self.json_second, fmt=Config.JSON)
        self.check_structure(c)



import unittest

from tea import ctx


class TestSuppress(unittest.TestCase):
    def test_raises_one(self):
        with ctx.suppress(IndexError):
            raise IndexError("Test")

    def test_raises_multiple(self):
        with ctx.suppress(IndexError, IOError):
            raise IndexError("Test")
        with ctx.suppress(IndexError, IOError):
            raise IOError("Test")

    def test_unexpected(self):
        with self.assertRaises(KeyError):
            with ctx.suppress(IndexError):
                raise KeyError("Test")
        with self.assertRaises(KeyError):
            with ctx.suppress(IndexError, IOError):
                raise KeyError("Test")

    def test_inheritance(self):
        with ctx.suppress(Exception):
            raise KeyError("Test")

        class MyError(KeyError):
            pass

        with ctx.suppress(KeyError):
            raise MyError("Test")



import logging
from re import sub
from html import parser


logger = logging.getLogger(__name__)


class _StripTagsParser(parser.HTMLParser):
    def __init__(self):
        super(_StripTagsParser, self).__init__()
        self.__text = []

    def handle_data(self, data):
        text = data.strip()
        if len(text) > 0:
            text = sub("[ \t\r\n]+", " ", text)
            self.__text.append(text + " ")

    def handle_starttag(self, tag, attrs):
        if tag == "p":
            self.__text.append("\n\n")
        elif tag == "br":
            self.__text.append("\n")

    def handle_startendtag(self, tag, attrs):
        if tag == "br":
            self.__text.append("\n\n")

    def text(self):
        return "".join(self.__text).strip()


def strip_tags(html):
    try:
        p = _StripTagsParser()
        p.feed(html)
        p.close()
        return p.text()
    except Exception as e:
        logger.debug("Failed to strip tags. Error: %s", e)
        return html



import os
import io
import json
import copy
import logging
import functools
import threading
from tea import shell


logger = logging.getLogger(__name__)


def locked(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        with self.lock:
            return func(self, *args, **kwargs)

    return wrapper


class Config(object):
    """Configuration class."""

    DICT = "dict"
    JSON = "json"

    def __init__(
        self,
        filename=None,
        data=None,
        fmt=None,
        encoding="utf-8",
        auto_save=True,
    ):
        self.lock = threading.Lock()
        self.encoding = encoding
        self.auto_save = auto_save
        if filename is not None:
            self.filename = os.path.abspath(filename)
            self.fmt = (
                fmt
                if fmt is not None
                else os.path.splitext(self.filename)[-1].lower()[1:]
            )
            self.data = self._read_file()
        else:
            self.filename = filename
            self.fmt = Config.JSON if fmt is None else fmt
            if isinstance(data, dict):
                self.fmt = Config.DICT
                self.data = copy.deepcopy(data)
            elif isinstance(data, (str, bytes)):
                self.data = self._read_string(data)
            else:
                self.data = {}

    def _read_file(self):
        if not os.path.isfile(self.filename):
            logger.warning(
                'Configuration file "%s" does not exist', self.filename
            )
            return {}
        try:
            if self.fmt == Config.JSON:
                with io.open(self.filename, "r", encoding=self.encoding) as f:
                    return json.loads(f.read())
            else:
                logger.error("Unsupported configuration format: %s", self.fmt)
                return {}
        except Exception as e:
            logger.error(
                'Failed to load file "%s" in format "%s". %s',
                self.filename,
                self.fmt,
                e,
            )
            return {}

    def _read_string(self, data):
        try:
            if isinstance(data, bytes):
                data = data.decode(self.encoding)
            if self.fmt == Config.JSON:
                return json.loads(data)
            else:
                logger.error("Unsupported configuration format: %s", self.fmt)
                return {}
        except Exception as e:
            logger.error('Failed to load data in format "%s". %s', self.fmt, e)
            return {}

    def save(self):
        if self.filename is not None:
            dirname = os.path.abspath(os.path.dirname(self.filename))
            if not os.path.isdir(dirname):
                shell.mkdir(dirname)
            if self.fmt == Config.JSON:
                with io.open(self.filename, "w", encoding=self.encoding) as f:
                    f.write(json.dumps(self.data, indent=2))
            else:
                logger.error("Unsupported configuration format: %s", self.fmt)

    def __get(self, var, create=False):
        current = self.data
        for part in var.split("."):
            if isinstance(current, dict):
                if part in current:
                    current = current[part]
                elif create:
                    current = current.setdefault(part, {})
                else:
                    raise KeyError(var)
            elif isinstance(current, list):
                try:
                    part = int(part, 10)
                    current = current[part]
                except Exception:
                    raise IndexError(var)
            else:
                raise KeyError(var)
        return current

    def __set(self, var, value):
        if "." in var:
            path, _, item = var.rpartition(".")
            current = self.__get(path, create=True)
        else:
            current = self.data
            item = var
        if isinstance(current, dict):
            current[item] = value
        elif isinstance(current, list):
            item = int(item, 10)
            current[item] = value
        if self.auto_save:
            self.save()

    def __del(self, var):
        if "." in var:
            path, _, item = var.rpartition(".")
            current = self.__get(path)
        else:
            current = self.data
            item = var
        if isinstance(current, dict):
            del current[item]
        elif isinstance(current, list):
            item = int(item, 10)
            current.pop(item)
        if self.auto_save:
            self.save()

    @locked
    def keys(self):
        """Return a set of top level keys in this configuration."""
        return set(self.data.keys())

    @locked
    def __getitem__(self, item):
        """Return a value from configuration.

        Unsafe version, may raise KeyError or IndexError.
        """
        return self.__get(item)

    @locked
    def __setitem__(self, item, value):
        return self.__set(item, value)

    @locked
    def __delitem__(self, item):
        """Delete an item from configuration.

        Unsafe version, may raise KeyError or IndexError.
        """
        return self.__del(item)

    @locked
    def __contains__(self, item):
        try:
            self.__get(item)
            return True
        except (KeyError, IndexError):
            return False

    @locked
    def get(self, var, default=None):
        """Return a value from configuration.

        Safe version which always returns a default value if the value is not
        found.
        """
        try:
            return self.__get(var)
        except (KeyError, IndexError):
            return default

    @locked
    def set(self, var, value):
        return self.__set(var, value)

    @locked
    def delete(self, var):
        """Delete an item from configuration.

        Safe version, never, raises an error.
        """
        try:
            return self.__del(var)
        except Exception:
            pass

    @locked
    def insert(self, var, value, index=None):
        """Insert at the index.

        If the index is not provided appends to the end of the list.
        """
        current = self.__get(var)
        if not isinstance(current, list):
            raise KeyError("%s: is not a list" % var)
        if index is None:
            current.append(value)
        else:
            current.insert(index, value)
        if self.auto_save:
            self.save()

    def __repr__(self):
        return (
            'Config(filename="%(filename)s", format="%(fmt)s", '
            'encoding="%(encoding)s", auto_save=%(auto_save)s)' % self.__dict__
        )


class MultiConfig(object):
    """Base class for configuration management."""

    def __init__(
        self,
        filename=None,
        data=None,
        fmt=None,
        encoding="utf-8",
        auto_save=True,
    ):
        self.lock = threading.Lock()
        self.__configs = []
        self.attach(filename, data, fmt, encoding, auto_save)

    @locked
    def attach(
        self,
        filename=None,
        data=None,
        fmt=None,
        encoding="utf-8",
        auto_save=True,
        index=None,
    ):
        config = Config(filename, data, fmt, encoding, auto_save)
        if index is None:
            self.__configs.insert(0, config)
        else:
            self.__configs.insert(index, config)

    @locked
    def detach(self, index=None):
        if index is None:
            self.__configs.pop(0)
        else:
            self.__configs.pop(index)

    @property
    def current(self):
        return self.__configs[0]

    def __get(self, var):
        for config in self.__configs:
            try:
                return config[var]
            except (KeyError, IndexError):
                pass
        raise KeyError(var)

    def __set(self, var, value):
        return self.current.set(var, value)

    def __del(self, var):
        # It has to keep track if it found a value in any of the configuration
        # files. If the value is found it won't raise and error, if it is not
        # found in any configuration file, it will raise a KeyError
        found = []
        for config in self.__configs:
            try:
                del config[var]
                found.append(True)
            except (KeyError, IndexError):
                found.append(True)
        if not any(found):
            raise KeyError(var)

    @locked
    def keys(self):
        """Return a merged set of top level keys from all configurations."""
        s = set()
        for config in self.__configs:
            s |= config.keys()
        return s

    @locked
    def __getitem__(self, item):
        """Return a value from configuration.

        Unsafe version, may raise KeyError or IndexError.
        """
        return self.__get(item)

    @locked
    def __setitem__(self, item, value):
        return self.__set(item, value)

    @locked
    def __delitem__(self, item):
        """Delete an item from configuration.

        Unsafe version, may raise KeyError or IndexError.
        """
        self.__del(item)

    @locked
    def __contains__(self, item):
        try:
            self.__get(item)
            return True
        except (KeyError, IndexError):
            return False

    @locked
    def get(self, var, default=None):
        """Return a value from configuration.

        Safe version which always returns a default value if the value is not
        found.
        """
        try:
            return self.__get(var)
        except (KeyError, IndexError):
            return default

    @locked
    def set(self, var, value):
        return self.__set(var, value)

    @locked
    def delete(self, var):
        """Delete an item from configuration.

        Safe version, never, raises an error.
        """
        try:
            self.__del(var)
        except (KeyError, IndexError):
            pass

    @locked
    def insert(self, var, value, index=None):
        self.current.insert(var, value, index)

    def __repr__(self):
        return "MultiConfig(\n  %s\n)" % (
            ",\n  ".join(reversed(map(repr, self.__configs)))
        )



import os
import logging
from functools import cmp_to_key
from typing import List, Generator, Tuple, Union, Dict, Optional

import psutil

from tea.utils import cmp
from tea.process.process import Process


logger = logging.getLogger(__name__)


def _list_processes() -> Generator:
    for p in psutil.process_iter():
        try:
            try:
                cmdline = p.cmdline()
            except Exception:
                cmdline = [p.exe()]
            yield Process.immutable(p.pid, cmdline)
        except Exception:
            pass


def get_processes(sort_by_name: bool = True) -> List[Process]:
    """Retrieve a list of processes sorted by name.

    Args:
        sort_by_name: Sort the list by name or by process ID's.

    Returns:
        List[Process]: List of processes.
    """
    if sort_by_name:
        return sorted(
            _list_processes(),
            key=cmp_to_key(
                lambda p1, p2: (
                    cmp(p1.command, p2.command) or cmp(p1.pid, p2.pid)
                )
            ),
        )
    else:
        return sorted(
            _list_processes(),
            key=cmp_to_key(
                lambda p1, p2: (
                    cmp(p1.pid, p2.pid) or cmp(p1.command, p2.command)
                )
            ),
        )


def find(name: str, arg: Optional[str] = None):
    """Find process by name or by argument in command line.

    Args:
        name: Process name to search for.
        arg: Command line argument for a process to search for.

    Returns:
        tea.process.base.IProcess: Process object if found.
    """
    for p in get_processes():
        if p.command.lower().find(name.lower()) != -1:
            if arg is not None:
                for a in p.command_line or []:
                    if a.lower().find(arg.lower()) != -1:
                        return p
            else:
                return p
    return None


def execute(
    command: Union[str, List[str]],
    env: Optional[Dict[str, str]] = None,
    working_dir: Optional[str] = None,
    wait: bool = True,
) -> Union[Tuple[int, str, str], Process]:
    """Execute a command with arguments and wait for output.

    Arguments should not be quoted!

    Keyword Arguments:
        command: Command to execute.
        env: Environment variables.
        working_dir: Set the working dir.
        wait: Wait for the process to finish.

    Returns:
        Tuple[int, str, str]: (exit_code, stdout, stderr) if wait is `True`
            else the process instance.

    Example::

        >>> code = 'import sys; sys.stdout.write('out'); sys.exit(0)'
        >>> status, out, err = execute(['python', '-c', code])
        >>> print('status: %s, output: %s, error: %s' % (status, out, err))
        status: 0, output: out, error:
        >>> code = 'import sys; sys.stderr.write('out'); sys.exit(1)'
        >>> status, out, err = execute(['python', '-c', code])
        >>> print('status: %s, output: %s, error: %s' % (status, out, err))
        status: 1, output: , error: err
    """
    process = Process(command=command, env=env, working_dir=working_dir)
    process.start()
    if not wait:
        return process
    process.wait()
    return process.exit_code, process.read(), process.eread()


def execute_no_demux(
    command: Union[str, List[str]],
    env: Optional[Dict[str, str]] = None,
    working_dir: Optional[str] = None,
    wait: bool = True,
) -> Union[Tuple[int, str], Process]:
    """Execute a command and return the exit code and output.

    Args:
        command: Command to execute.
        env: Environment variables.
        working_dir: Set the working dir.
        wait: Wait for the process to finish.

    Returns:
        Tuple[int, str, str]: (exit_code, stdout, stderr) if wait is `True`
            else the process instance.
    """
    process = Process(
        command=command, env=env, working_dir=working_dir, demux=False
    )
    process.start()
    if not wait:
        return process
    process.wait()
    return process.exit_code, process.read()


def execute_and_report(command, *args, **kwargs):
    """Execute a command with arguments and wait for output.

    If execution was successful function will return True,
    if not, it will log the output using standard logging and return False.
    """
    logging.info("Execute: %s" % command)
    try:
        status, out, err = execute(command, *args, **kwargs)
        if status == 0:
            logging.info(
                "%s Finished successfully. Exit Code: 0.",
                os.path.basename(command[0]),
            )
            return True
        else:
            try:
                logging.error(
                    "%s failed! Exit Code: %s\nOut: %s\nError: %s",
                    os.path.basename(command[0]),
                    status,
                    out,
                    err,
                )
            except Exception as e:
                # This fails when some non ASCII characters are returned
                # from the application
                logging.error(
                    "%s failed [%s]! Exit Code: %s\nOut: %s\nError: %s",
                    e,
                    os.path.basename(command[0]),
                    status,
                    repr(out),
                    repr(err),
                )
            return False
    except Exception:
        logging.exception(
            "%s failed! Exception thrown!", os.path.basename(command[0])
        )
        return False



import pytest
from tea.decorators import combomethod, ComboMethodError


def test_static_and_instance_method_combo():
    class Foo(object):
        def __init__(self):
            self.instance_variable = 2

        @combomethod(static=True)
        def static_and_instance(x):
            return x + 1

        @static_and_instance.instance
        def static_and_instance(self, x):
            return x + self.instance_variable

    assert Foo.static_and_instance(100) == 101
    assert Foo().static_and_instance(100) == 102


def test_class_and_instance_method_combo():
    class Foo(object):
        class_variable = 1

        def __init__(self):
            self.instance_variable = 2
            # Override class variable for test case
            self.class_variable = 3

        @combomethod
        def class_and_instance(cls, x):
            return x + cls.class_variable

        @class_and_instance.instance
        def class_and_instance(self, x):
            return x + self.instance_variable

    assert Foo.class_and_instance(100) == 101
    assert Foo().class_and_instance(100) == 102


def test_combomethod_error():
    class Foo(object):
        @combomethod
        def foo(cls):
            return 0

    assert Foo.foo() == 0
    pytest.raises(ComboMethodError, lambda: Foo().foo())

    class Bar(object):
        @combomethod(static=True)
        def bar():
            return 0

    assert Bar.bar() == 0
    pytest.raises(ComboMethodError, lambda: Bar().bar())



import sys


def is_true(x):
    return x.lower() in ("true", "t")


commands = {
    # Function name : (Iterable result, [Param types],
    #                  Number of required arguments)])
    "search": (True, [str, str, is_true, is_true], 2),
    "chdir": (False, [str], 1),
    "mkdir": (False, [str], 1),
    "copy": (False, [str, str], 2),
    "gcopy": (False, [str, str], 2),
    "move": (False, [str, str], 2),
    "gmove": (False, [str, str], 2),
    "remove": (False, [str], 1),
    "gremove": (False, [str], 1),
}


def print_help():
    print(
        "Usage: python -m tea.shell [command] [params]\nCommands: %s"
        % ", ".join(commands.keys())
    )
    return 1


def print_func_help(func, required):
    import inspect

    args = inspect.getargspec(func).args
    print("Usage: %s" % func.__name__, end=" ")
    for i in range(required):
        print(args[i], end=" ")
    if required < len(args):
        print("[%s]" % " ".join(args[required:]))
    return 1


def main(args):
    # Configure logger
    import logging
    from tea.logger import configure_logging

    configure_logging(stdout_level=logging.WARNING)

    if len(args) == 0:
        print_help()
    else:
        command = args[0]
        if command == "help" or command not in commands:
            return print_help()
        else:
            args = args[1:]
            # get specs
            iterable, params, required = commands[command]
            # import function
            from tea import shell

            func = getattr(shell, command)
            if not (required <= len(args) <= len(params)):
                return print_func_help(func, required)
            parsed_args = []
            for i, arg in enumerate(args):
                try:
                    parsed_args.append(params[i](arg))
                except Exception:
                    print("Failed to parse argument: %s" % arg)
                    return 1
            try:
                if iterable:
                    for i in func(*parsed_args):
                        print(i)
                else:
                    if func(*parsed_args):
                        print("OK")
                        return 0
                    else:
                        print("FAILED")
                        return 1
            except Exception as e:
                print("ERROR: %s" % e)
                return 1
            return 0


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



def smart_text(s, encoding="utf-8", errors="strict"):
    """Return a unicode object representing 's'.

    Treats bytes using the 'encoding' codec.
    """
    if isinstance(s, str):
        return s

    if isinstance(s, bytes):
        return s.decode(encoding, errors)

    return str(s)


def smart_bytes(s, encoding="utf-8", errors="strict"):
    """Return a bytes version of 's' encoded as specified in 'encoding'."""
    if isinstance(s, bytes):
        if encoding == "utf-8":
            return s
        else:
            return s.decode("utf-8", errors).encode(encoding, errors)

    if isinstance(s, str):
        return s.encode(encoding, errors)

    return str(s).encode(encoding, errors)



class TeaError(Exception):
    """Base class for all errors.

    This error class will be used as a base class for all other modules
    that use tea as it's base.
    """

    def __init__(self, message: str):
        self.message = message

    def __str__(self):
        return f"{self.__class__.__name__}({self.message})"

    __repr__ = __str__



import tempfile
import unittest
import logging
from tea.logger import configure_logging


DATE = r"\d{4}\.\d{2}\.\d{2} \d{2}:\d{2}:\d{2}.\d{3}"


class TestLogger(unittest.TestCase):
    def setUp(self):
        self.output_filename = tempfile.mktemp()
        configure_logging(self.output_filename, level=logging.DEBUG)
        self.logger = logging.getLogger(__name__)
        self.output_test_file = open(self.output_filename)

    def tearDown(self):
        self.output_test_file.close()
        # TODO: You cannot delete open file on windows, and the file
        # remains open
        # because the logging is holding it.
        # os.remove(self.output_filename)

    def test_debug(self):
        self.logger.debug("Debug message #01")
        regexp = DATE + r"\s+ DEBUG: Debug message #01 \[[\w\.]+:\d{2,3}\]"
        self.assertRegex(self.output_test_file.readline(), regexp)

    def test_info(self):
        self.logger.info("Info message #02")
        regexp = DATE + r"\s+ INFO: Info message #02 \[[\w\.]+:\d{2,3}\]"
        self.assertRegex(self.output_test_file.readline(), regexp)

    def test_warning(self):
        self.logger.warning("Warning message #03")
        regexp = DATE + r"\s+ WARNING: Warning message #03 \[[\w\.]+:\d{2,3}\]"
        self.assertRegex(self.output_test_file.readline(), regexp)

    def test_error(self):
        self.logger.error("Error message #04")
        regexp = DATE + r"\s+ ERROR: Error message #04 \[[\w\.]+:\d{2,3}\]"
        self.assertRegex(self.output_test_file.readline(), regexp)

    def test_critical(self):
        self.logger.critical("Critical message #06")
        regexp = (
            DATE + r"\s+ CRITICAL: Critical message #06 \[[\w\.]+:\d{2,3}\]"
        )
        self.assertRegex(self.output_test_file.readline(), regexp)

    def test_exception(self):
        try:
            raise Exception("Error")
        except Exception:
            self.logger.exception("Exception #08")
        regexp = DATE + r"\s+ ERROR: Exception #08 \[[\w\.]+:\d{2,3}\]"
        self.assertRegex(self.output_test_file.readline(), regexp)
        traceback = self.output_test_file.read().strip()
        self.assertTrue(
            traceback.startswith("Traceback (most recent call last):")
        )
        self.assertTrue(traceback.endswith("Exception: Error"))



import types
import pkgutil
import logging
from tea.utils import get_object


logger = logging.getLogger(__name__)


def anotate(name):
    """Annotate an object with a name."""

    def decorator(obj):
        setattr(obj, "_tea_ds_plugin", name)
        return obj

    return decorator


def load_plugins(modules, cls=None, annotation=None, subclasses=False):
    if cls is None and annotation is None:
        logger.warning(
            "Either the cls or the annotation has to be provided "
            "to the load_plugins function."
        )
        return []
    if not isinstance(modules, (list, tuple)):
        modules = [modules]
    # First load all the modules
    loaded_modules = []
    for module in modules:
        if not isinstance(module, types.ModuleType):
            module = get_object(module)
        loaded_modules.append(module)
        for (loader, module_name, is_pkg) in pkgutil.walk_packages(
            module.__path__
        ):
            loaded_modules.append(
                loader.find_module(module_name).load_module(module_name)
            )
    # If cls is provided then check the classes
    if cls is not None:
        if subclasses:
            classes = []

            def _get_sublasses(cls):
                for c in cls.__subclasses__():
                    classes.append(c)
                    _get_sublasses(c)

            _get_sublasses(cls)
            return classes
        else:
            return cls.__subclasses__()
    # If annotation is provided then check the annotations
    if annotation is not None:
        items = []
        for module in loaded_modules:
            for obj in module.__dict__.values():
                if getattr(obj, "_tea_ds_plugin", None) == annotation:
                    items.append(obj)
        return items



__all__ = ["combomethod", "ComboMethodError"]

import functools


class ComboMethodError(Exception):
    """Inappropriate initialization of the combo method descriptor."""

    pass


class ComboMethod(object):
    """Combo method descriptor.

    Descriptor is initializes eater by passing in the staticmethod or the
    class method. After that an instance method can be added using the instance
    decorator.

    This class should never be used. Use the ``combomethod`` function instead.
    """

    def __init__(self, staticmethod=None, classmethod=None):
        if staticmethod is None and classmethod is None:
            raise ComboMethodError(
                "Either static method or class method has to be provided"
            )
        self.staticmethod = staticmethod
        self.classmethod = classmethod
        self.instancemethod = None

    def instance(self, instancemethod):
        self.instancemethod = instancemethod
        return self

    def __get__(self, obj, type=None):
        if obj is None:
            if self.staticmethod is not None:
                return self.staticmethod
            else:

                @functools.wraps(self.classmethod)
                def wrapper(*args, **kwargs):
                    return self.classmethod(type, *args, **kwargs)

                return wrapper
        else:
            if self.instancemethod is None:
                raise ComboMethodError("Instance method is not provided")

            @functools.wraps(self.instancemethod)
            def wrapper(*args, **kwargs):
                return self.instancemethod(obj, *args, **kwargs)

            return wrapper


def combomethod(method=None, static=False):
    """Create a class method or static method.

    It will be used when you call it on the class but can be overridden by an
    instance method of the same name that will be called when the method is
    called on the instance.

    Usage::

        class Foo(object):
            class_variable = 2

            def __init__(self):
                self.instance_variable = 3
                # Override class variable for test case
                self.class_variable = 4

            @combomethod(static=True)
            def static_and_instance(x):
                return x + 1

            @static_and_instance.instance
            def static_and_instance(self, x):
                return x + self.instance_variable

            @combomethod
            def class_and_instance(cls, x):
                return x + cls.class_variable

            @class_and_instance.instance
            def class_and_instance(self, x):
                return x + self.instance_variable

        >>> Foo.static_and_instance(100)
        101
        >>> Foo.class_and_instance(100)
        102
        >>> f = Foo()
        >>> f.static_and_instance(100)
        103
        >>> f.class_and_instance(100)
        103
    """
    return ComboMethod if static else ComboMethod(None, method)



import enum
import uuid
import decimal
import datetime
from dataclasses import dataclass, asdict

import pytz


from tea import serde
from tea import timestamp as ts


class MyEnum(enum.Enum):
    foo = "foo"
    bar = "bar"


@dataclass
class DC:
    x: int
    y: str


class Foo:
    def __init__(self, x, y):
        self.x = x
        self.y = y

    @property
    def sum(self):
        return self.x + self.y

    def to_dict(self):
        return {"x": self.x, "y": self.y, "sum": self.sum}


def test_json_encode():
    d = {"a": 1, "b": 2}
    foo = Foo(3, 4)
    uid = uuid.uuid4()
    dt = datetime.datetime(2017, 3, 16, 18, 30, 15)
    date = datetime.date(2017, 3, 16)
    dc = DC(x=3, y="foo")
    timezone = pytz.timezone("Europe/London")

    data = serde.json_loads(
        serde.json_dumps(
            {
                "to_dict": foo,
                "decimal": decimal.Decimal("3.0"),
                "datetime": dt,
                "date": date,
                "set": {1, 2, 3},
                "dict_keys": d.keys(),
                "dict_values": d.values(),
                "uuid": uid,
                "enum": MyEnum.foo,
                "dataclass": dc,
                "timezone": timezone,
            }
        )
    )
    assert data["to_dict"] == foo.to_dict()
    assert data["decimal"] == 3.0
    assert data["datetime"] == ts.to_utc_str(dt)
    assert data["date"] == ts.to_utc_str(date)
    assert set(data["set"]) == {1, 2, 3}
    assert set(data["dict_keys"]) == set(d.keys())
    assert set(data["dict_values"]) == set(d.values())
    assert data["uuid"] == str(uid)
    assert data["enum"] == MyEnum.foo.value
    assert data["dataclass"] == asdict(dc)
    assert data["timezone"] == timezone.zone



"""Simple and complete library for sending emails."""

__all__ = [
    "SMTPConnection",
    "EmailMessage",
    "EmailMultiAlternatives",
    "send_mail",
    "send_mass_mail",
]


import os
import time
import socket
import random
import smtplib
import logging
import mimetypes
from email import charset as Charset
from email.mime.text import MIMEText
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.header import Header
from email.encoders import encode_base64
from email.utils import formatdate, parseaddr, formataddr

from tea.utils.html import strip_tags
from tea.utils.encoding import smart_text, smart_bytes


logger = logging.getLogger(__name__)


# Don't BASE64-encode UTF-8 messages so that we avoid unwanted attention from
# some spam filters.
Charset.add_charset("utf-8", Charset.SHORTEST, Charset.QP, "utf-8")

# Default MIME type to use on attachments (if it is not explicitly given
# and cannot be guessed).
DEFAULT_ATTACHMENT_MIME_TYPE = "application/octet-stream"

# Default charset
DEFAULT_CHARSET = "utf-8"


# Cache the hostname, but do it lazily: socket.getfqdn() can take a couple of
# seconds, which slows down the restart of the server.
class CachedDnsName(object):
    def __str__(self):
        return self.get_fqdn()

    def get_fqdn(self):
        if not hasattr(self, "_fqdn"):
            self._fqdn = socket.getfqdn()
        return self._fqdn


DNS_NAME = CachedDnsName()


# Copied from Python standard library, with the following modifications:
# * Used cached hostname for performance.
# * Added try/except to support lack of getpid() in Jython (#5496).
def make_msgid(idstring=None, utc=False):
    """Return a string suitable for RFC 2822 compliant Message-ID.

    E.g: <20020201195627.33539.96671@nightshade.la.mastaler.com>

    Optional idstring if given is a string used to strengthen the
    uniqueness of the message id.
    """
    if utc:
        timestamp = time.gmtime()
    else:
        timestamp = time.localtime()
    utcdate = time.strftime("%Y%m%d%H%M%S", timestamp)
    try:
        pid = os.getpid()
    except AttributeError:
        # No getpid() in Jython, for example.
        pid = 1
    randint = random.randrange(100000)
    if idstring is None:
        idstring = ""
    else:
        idstring = "." + idstring
    idhost = DNS_NAME
    msgid = "<%s.%s.%s%s@%s>" % (utcdate, pid, randint, idstring, idhost)
    return msgid


class BadHeaderError(ValueError):
    pass


def forbid_multi_line_headers(name, val):
    """Forbid multi-line headers, to prevent header injection."""
    val = smart_text(val)
    if "\n" in val or "\r" in val:
        raise BadHeaderError(
            "Header values can't contain newlines "
            "(got %r for header %r)" % (val, name)
        )
    try:
        val = val.encode("ascii")
    except UnicodeEncodeError:
        if name.lower() in ("to", "from", "cc"):
            result = []
            for item in val.split(", "):
                nm, addr = parseaddr(item)
                nm = str(Header(nm, DEFAULT_CHARSET))
                result.append(formataddr((nm, str(addr))))
            val = ", ".join(result)
        else:
            val = Header(val, DEFAULT_CHARSET)
    else:
        if name.lower() == "subject":
            val = Header(val)
    return name, val


class SafeMIMEText(MIMEText):
    def __setitem__(self, name, val):
        name, val = forbid_multi_line_headers(name, val)
        MIMEText.__setitem__(self, name, val)


class SafeMIMEMultipart(MIMEMultipart):
    def __setitem__(self, name, val):
        name, val = forbid_multi_line_headers(name, val)
        MIMEMultipart.__setitem__(self, name, val)


class SMTPConnection(object):
    """Wrapper that manages the SMTP network connection."""

    def __init__(
        self,
        host=None,
        port=None,
        username=None,
        password=None,
        use_tls=None,
        fail_silently=False,
    ):
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.use_tls = (use_tls is not None) and use_tls
        self.fail_silently = fail_silently
        self.connection = None

    def open(self):
        """Ensure we have a connection to the email server.

        Returns whether or not a new connection was required (True or False).
        """
        if self.connection:
            # Nothing to do if the connection is already open.
            return False
        try:
            # If local_hostname is not specified, socket.getfqdn() gets used.
            # For performance, we use the cached FQDN for local_hostname.
            self.connection = smtplib.SMTP(
                self.host, self.port, local_hostname=DNS_NAME.get_fqdn()
            )
            if self.use_tls:
                self.connection.ehlo()
                self.connection.starttls()
                self.connection.ehlo()
            if self.username and self.password:
                self.connection.login(self.username, self.password)
            return True
        except Exception as e:
            logger.error(
                "Error trying to connect to server %s:%s: %s",
                self.host,
                self.port,
                e,
            )
            if not self.fail_silently:
                raise

    def close(self):
        """Close the connection to the email server."""
        try:
            try:
                self.connection.quit()
            except socket.sslerror:
                # This happens when calling quit() on a TLS connection
                # sometimes.
                self.connection.close()
            except Exception as e:
                logger.error(
                    "Error trying to close connection to server " "%s:%s: %s",
                    self.host,
                    self.port,
                    e,
                )
                if self.fail_silently:
                    return
                raise
        finally:
            self.connection = None

    def send_messages(self, messages):
        """Send one or more EmailMessage objects.

        Returns:
             int: Number of email messages sent.
        """
        if not messages:
            return
        new_conn_created = self.open()
        if not self.connection:
            # We failed silently on open(). Trying to send would be pointless.
            return
        num_sent = 0
        for message in messages:
            sent = self._send(message)
            if sent:
                num_sent += 1
        if new_conn_created:
            self.close()
        return num_sent

    def _send(self, message):
        """Send an email.

        Helper method that does the actual sending.
        """
        if not message.recipients():
            return False
        try:
            self.connection.sendmail(
                message.sender,
                message.recipients(),
                message.message().as_string(),
            )
        except Exception as e:
            logger.error(
                "Error sending a message to server %s:%s: %s",
                self.host,
                self.port,
                e,
            )
            if not self.fail_silently:
                raise
            return False
        return True


class EmailMessage(object):
    """Container for email information."""

    content_subtype = "plain"
    multipart_subtype = "mixed"
    encoding = None  # None => use settings default

    def __init__(
        self,
        subject="",
        body="",
        sender=None,
        to=None,
        cc=None,
        bcc=None,
        attachments=None,
        headers=None,
        connection=None,
    ):
        """Initialize a single email message.

        (It can be sent to multiple recipients.)

        All strings used to create the message can be unicode strings (or UTF-8
        bytestrings). The SafeMIMEText class will handle any necessary encoding
        conversions.
        """
        if to:
            if not isinstance(to, (list, tuple)):
                raise AssertionError('"to" argument must be a list or tuple')
            self.to = list(to)
        else:
            self.to = []
        if cc:
            if not isinstance(cc, (list, tuple)):
                raise AssertionError('"cc" argument must be a list or tuple')
            self.cc = list(cc)
        else:
            self.cc = []
        if bcc:
            if not isinstance(bcc, (list, tuple)):
                raise AssertionError('"bcc" argument must be a list or tuple')
            self.bcc = list(bcc)
        else:
            self.bcc = []
        self.sender = sender
        self.subject = subject
        self.body = body
        self.attachments = []
        for attachment in attachments or []:
            if isinstance(attachment, (tuple, list)):
                self.attach(*attachment)
            else:
                self.attach(attachment)
        self.extra_headers = headers or {}
        self.connection = connection

    def get_connection(self, fail_silently=False):
        if not self.connection:
            self.connection = SMTPConnection(fail_silently=fail_silently)
        return self.connection

    def message(self):
        encoding = self.encoding or DEFAULT_CHARSET
        msg = SafeMIMEText(
            smart_bytes(self.body, DEFAULT_CHARSET),
            self.content_subtype,
            encoding,
        )
        if self.attachments:
            body_msg = msg
            msg = SafeMIMEMultipart(_subtype=self.multipart_subtype)
            if self.body:
                msg.attach(body_msg)
            for attachment in self.attachments:
                if isinstance(attachment, MIMEBase):
                    msg.attach(attachment)
                else:
                    msg.attach(self._create_attachment(*attachment))
        msg["Subject"] = self.subject
        msg["From"] = self.sender
        msg["To"] = ", ".join(self.to)
        msg["Cc"] = ", ".join(self.cc)
        # Email header names are case-insensitive (RFC 2045), so we have to
        # accommodate that when doing comparisons.
        header_names = [key.lower() for key in self.extra_headers]
        if "date" not in header_names:
            msg["Date"] = formatdate()
        if "message-id" not in header_names:
            msg["Message-ID"] = make_msgid()
        for name, value in self.extra_headers.items():
            msg[name] = value
        return msg

    def recipients(self):
        """Return a list of all recipients of the email.

        Includes direct addressees as well as Bcc entries.
        """
        return self.to + self.cc + self.bcc

    def send(self, fail_silently=False):
        """Send the email message."""
        return self.get_connection(fail_silently).send_messages([self])

    def attach(self, filename=None, content=None, mimetype=None):
        """Attache a file with the given filename and content.

        The filename can be omitted (useful for multipart/alternative messages)
        and the mimetype is guessed, if not provided.

        If the first parameter is a MIMEBase subclass it is inserted directly
        into the resulting message attachments.
        """
        if isinstance(filename, MIMEBase):
            assert content is None and mimetype is None
            self.attachments.append(filename)
        elif content is None and os.path.isfile(filename):
            self.attach_file(filename, mimetype)
        else:
            assert content is not None
            self.attachments.append((filename, content, mimetype))

    def attach_file(self, path, mimetype=None):
        """Attache a file from the filesystem."""
        filename = os.path.basename(path)
        content = open(path, "rb").read()
        self.attach(filename, content, mimetype)

    def _create_attachment(self, filename, content, mimetype=None):
        """Convert the filename, content, mimetype triple to attachment."""
        if mimetype is None:
            mimetype, _ = mimetypes.guess_type(filename)
            if mimetype is None:
                mimetype = DEFAULT_ATTACHMENT_MIME_TYPE
        basetype, subtype = mimetype.split("/", 1)
        if basetype == "text":
            attachment = SafeMIMEText(
                smart_bytes(content, DEFAULT_CHARSET), subtype, DEFAULT_CHARSET
            )
        else:
            # Encode non-text attachments with base64.
            attachment = MIMEBase(basetype, subtype)
            attachment.set_payload(content)
            encode_base64(attachment)
        if filename:
            attachment.add_header(
                "Content-Disposition", "attachment", filename=filename
            )
        return attachment


class EmailMultiAlternatives(EmailMessage):
    """Email message with multipart/alternative content.

    A version of EmailMessage that makes it easy to send
    multipart/alternative messages. For example, including text and HTML
    versions of the text is made easier.
    """

    multipart_subtype = "alternative"

    def attach_alternative(self, content, mimetype=None):
        """Attach an alternative content representation."""
        self.attach(content=content, mimetype=mimetype)


def send_mail(
    subject,
    sender,
    to,
    message,
    html_message=None,
    cc=None,
    bcc=None,
    attachments=None,
    host=None,
    port=None,
    auth_user=None,
    auth_password=None,
    use_tls=False,
    fail_silently=False,
):
    """Send a single email to a recipient list.

    All members of the recipient list will see the other recipients in the 'To'
    field.

    Note: The API for this method is frozen. New code wanting to extend the
    functionality should use the EmailMessage class directly.
    """
    if message is None and html_message is None:
        raise ValueError("Either message or html_message must be provided")
    if message is None:
        message = strip_tags(html_message)
    connection = SMTPConnection(
        host=host,
        port=port,
        username=auth_user,
        password=auth_password,
        use_tls=use_tls,
        fail_silently=fail_silently,
    )
    # Convert the to field just for easier usage
    if isinstance(to, str):
        to = [to]
    if html_message is None:
        email = EmailMessage(
            subject=subject,
            body=message,
            sender=sender,
            to=to,
            cc=cc,
            bcc=bcc,
            attachments=attachments,
            connection=connection,
        )
    else:
        email = EmailMultiAlternatives(
            subject=subject,
            body=message,
            sender=sender,
            to=to,
            cc=cc,
            bcc=bcc,
            attachments=attachments,
            connection=connection,
        )
        email.attach_alternative(html_message, "text/html")
    return email.send()


def send_mass_mail(
    datatuple, fail_silently=False, auth_user=None, auth_password=None
):
    """Send multiple emails to multiple recipients.

    Given a datatuple of (subject, message, sender, recipient_list), sends
    each message to each recipient list. Returns the number of e-mails sent.

    If auth_user and auth_password are set, they're used to log in.

    Note: The API for this method is frozen. New code wanting to extend the
    functionality should use the EmailMessage class directly.
    """
    connection = SMTPConnection(
        username=auth_user, password=auth_password, fail_silently=fail_silently
    )
    messages = [
        EmailMessage(subject, message, sender, recipient)
        for subject, message, sender, recipient in datatuple
    ]
    return connection.send_messages(messages)



"""Serialization and deserialization library."""

import enum
import json
import dataclasses

from tea import timestamp as ts


class TeaJsonEncoder(json.JSONEncoder):
    """Tea JSON Encoder.

    It knows how to serialize:

        1. All objects that have a custom `to_dict` method.
        2. Decimal numbers.
        3. DateTime, Date adn Timezone objects.
        4. Enums.
        5. Dataclasses
        6. UUIDs.
    """

    to_float = frozenset(("decimal.Decimal",))
    to_datetime = frozenset(("datetime.datetime", "datetime.date"))
    to_list = frozenset(
        (
            "__builtin__.set",
            "builtins.set",
            "builtins.dict_keys",
            "builtins.dict_values",
        )
    )
    to_str = frozenset(("uuid.UUID",))

    def default(self, o):
        try:
            return super(TeaJsonEncoder, self).default(o)
        except TypeError:
            # First see if there is a to_dict method
            if hasattr(o, "to_dict"):
                return o.to_dict()
            # Then try out special classes
            cls = o.__class__
            path = "%s.%s" % (cls.__module__, cls.__name__)
            if path in self.to_float:
                return float(o)
            elif path in self.to_datetime:
                return ts.to_utc_str(o)
            elif path in self.to_list:
                return list(o)
            elif path in self.to_str:
                return str(o)
            elif isinstance(o, enum.Enum):
                return o.value
            elif dataclasses.is_dataclass(o):
                return dataclasses.asdict(o)
            elif path.startswith("pytz.tzfile."):
                return o.zone
            raise TypeError("%s is not JSON serializable" % o)


json_loads = json.loads


def json_dumps(obj, indent=4) -> str:
    """Wrap `json.dumps` using the `TeaJsonEncoder`."""
    return json.dumps(
        obj,
        cls=TeaJsonEncoder,
        ensure_ascii=False,
        allow_nan=False,
        indent=indent,
        separators=(",", ":"),
    )



import unittest

from tea.utils import get_object


class Foo(object):
    def __init__(self, x, y=None):
        self.x = x
        self.y = y

    def foo(self):
        return self.x


class TestGetObject(unittest.TestCase):
    def setUp(self):
        self.x = 42
        self.foo = Foo(self.x)

    def test_object_getter(self):
        self.assertEqual(get_object("foo", self), self.foo)
        self.assertEqual(get_object("foo.x", self), self.x)
        self.assertEqual(get_object("foo.foo", self)(), self.x)

    def test_module_getter(self):
        self.assertEqual(get_object("unittest"), unittest)
        value = get_object("subprocess")
        import subprocess

        self.assertEqual(value, subprocess)
        self.assertEqual(get_object("tea.tests.test_utils.unittest"), unittest)

    def test_module_object_getter(self):
        self.assertEqual(get_object("unittest.TestCase"), unittest.TestCase)
        self.assertEqual(get_object("tea.utils.get_object"), get_object)
        self.assertEqual(get_object("tea.tests.test_utils.Foo.foo"), Foo.foo)



"""Logging module.

This logging module is designed as a wrapper around the python `logging`
module.

When the module is loaded it configure the logging object with some
default parameters (log to stderr with DEBUG level).
After that, user can call the L{configure_logger} function and
configure logging to file and stderr.


# Usage

>>> import logging
>>> import tempfile
>>> from tea.logger import configure_logging
>>> configure_logging(filename=tempfile.mktemp())
>>> logger = logging.getLogger('test')
>>> logger.debug('Debug level log entry')
>>> logger.info('Info level log entry')
>>> logger.warning('Warn level log entry')
WARNING     - Warn level log entry [test:1]
>>> logger.error('Error level log entry')
ERROR       - Error level log entry [test:1]
>>> logger.critical('Critical level log entry')
CRITICAL    - Critical level log entry [test:1]
>>> try:
...     raise Exception('Test exception')
... except Exception:
...     logger.exception('Error level log entry with stack trace')
ERROR       - Error level log entry with stack trace [test:4]
Traceback (most recent call last):
  ...
Exception: Test exception
"""

import os
import sys
import time
import logging
from logging.handlers import RotatingFileHandler

from tea import shell

# Constants
FMT = (
    "%(asctime)s.%(msecs)03d   %(levelname)11s: %(message)s "
    "[%(name)s:%(lineno)d]"
)
FMT_SHORT = "%(asctime)s %(levelname)11s: %(message)s [%(name)s:%(lineno)d]"
FMT_STDOUT = "%(levelname)-11s - %(message)s [%(name)s:%(lineno)d]"
FMT_TS = "%Y.%m.%d %H:%M:%S"

logging.basicConfig(
    stream=sys.stderr, format=FMT_SHORT, datefmt=FMT_TS, level=logging.DEBUG
)


def configure_logging(
    filename=None,
    filemode="a",
    ts_fmt=FMT_TS,
    fmt=FMT,
    stdout_fmt=FMT_STDOUT,
    level=logging.DEBUG,
    stdout_level=logging.WARNING,
    initial_file_message="",
    max_size=1048576,
    rotations_number=5,
    remove_handlers=True,
):
    """Configure logging module.

    Args:
        filename (str): Specifies a filename to log to.
        filemode (str): Specifies the mode to open the log file.
            Values: ``'a'``, ``'w'``. *Default:* ``a``.
        ts_fmt (str): Use the specified timestamp format.
        fmt (str): Format string for the file handler.
        stdout_fmt (str): Format string for the stdout handler.
        level (int): Log level for the file handler. Log levels are the same
            as the log levels from the standard :mod:`logging` module.
            *Default:* ``logging.DEBUG``
        stdout_level (int): Log level for the stdout handler. Log levels are
            the same as the log levels from the standard :mod:`logging` module.
            *Default:* ``logging.WARNING``
        initial_file_message (str): First log entry written in file.
        max_size (int): Maximal size of the logfile. If the size of the file
            exceed the maximal size it will be rotated.
        rotations_number (int): Number of rotations to save.
        remove_handlers (bool): Remove all existing handlers.
    """
    logger = logging.getLogger()
    logger.level = logging.NOTSET
    # Remove all handlers
    if remove_handlers:
        while len(logger.handlers) > 0:
            hdlr = logger.handlers[0]
            hdlr.close()
            logger.removeHandler(hdlr)
    # Create stdout handler
    if stdout_level is not None:
        stdout_handler = logging.StreamHandler(sys.stdout)
        stdout_handler.setLevel(stdout_level)
        stdout_formatter = logging.Formatter(stdout_fmt, ts_fmt)
        # stdoutFormatter.converter = time.gmtime
        stdout_handler.setFormatter(stdout_formatter)
        logger.addHandler(stdout_handler)
    # Create file handler if filename is provided
    if filename is not None:
        # Check if filename directory exists and creates it if it doesn't
        directory = os.path.abspath(os.path.dirname(filename))
        if not os.path.isdir(directory):
            shell.mkdir(directory)
        # Create file handler
        file_handler = RotatingFileHandler(
            filename, filemode, max_size, rotations_number
        )
        file_handler.setLevel(level)
        file_formatter = logging.Formatter(fmt, ts_fmt)
        file_formatter.converter = time.gmtime
        file_handler.setFormatter(file_formatter)
        logger.addHandler(file_handler)
        if initial_file_message:
            message = " %s " % initial_file_message
            file_handler.stream.write("\n" + message.center(100, "=") + "\n\n")



"""Module mimics some of the behaviors of the builtin :mod:`shutil`.

It adds logging to all operations and abstracting some other useful shell
commands (functions).
"""

import os
import io
import glob
import shlex
import shutil
import fnmatch
import logging
import contextlib


logger = logging.getLogger(__name__)


def split(s, posix=True):
    """Split the string s using shell-like syntax.

    Args:
        s (str): String to split
        posix (bool): Use posix split

    Returns:
        list of str: List of string parts
    """
    if isinstance(s, bytes):
        s = s.decode("utf-8")
    return shlex.split(s, posix=posix)


def search(path, matcher="*", dirs=False, files=True):
    """Recursive search function.

    Args:
        path (str): Path to search recursively
        matcher (str or callable): String pattern to search for or function
            that returns True/False for a file argument
        dirs (bool): if True returns directories that match the pattern
        files(bool): if True returns files that match the patter

    Yields:
        str: Found files and directories
    """
    if callable(matcher):

        def fnmatcher(items):
            return list(filter(matcher, items))

    else:

        def fnmatcher(items):
            return fnmatch.filter(items, matcher)

    for root, directories, filenames in os.walk(os.path.abspath(path)):
        to_match = []
        if dirs:
            to_match.extend(directories)
        if files:
            to_match.extend(filenames)
        for item in fnmatcher(to_match):
            yield os.path.join(root, item)


def chdir(directory):
    """Change the current working directory.

    Args:
        directory (str): Directory to go to.
    """
    directory = os.path.abspath(directory)
    logger.info("chdir -> %s" % directory)
    try:
        if not os.path.isdir(directory):
            logger.error(
                "chdir -> %s failed! Directory does not exist!", directory
            )
            return False
        os.chdir(directory)
        return True
    except Exception as e:
        logger.error("chdir -> %s failed! %s" % (directory, e))
        return False


@contextlib.contextmanager
def goto(directory, create=False):
    """Context object for changing directory.

    Args:
        directory (str): Directory to go to.
        create (bool): Create directory if it doesn't exists.

    Usage::

        >>> with goto(directory) as ok:
        ...     if not ok:
        ...         print 'Error'
        ...     else:
        ...         print 'All OK'
    """

    current = os.getcwd()
    directory = os.path.abspath(directory)

    if os.path.isdir(directory) or (create and mkdir(directory)):
        logger.info("goto -> %s", directory)
        os.chdir(directory)
        try:
            yield True
        finally:
            logger.info("goto <- %s", directory)
            os.chdir(current)
    else:
        logger.info(
            "goto(%s) - directory does not exist, or cannot be " "created.",
            directory,
        )
        yield False


def mkdir(path, mode=0o755, delete=False):
    """Make a directory.

    Create a leaf directory and all intermediate ones.
    Works like ``mkdir``, except that any intermediate path segment (not just
    the rightmost) will be created if it does not exist. This is recursive.

    Args:
        path (str): Directory to create
        mode (int): Directory mode
        delete (bool): Delete directory/file if exists

    Returns:
        bool: True if succeeded else False
    """
    logger.info("mkdir: %s" % path)
    if os.path.isdir(path):
        if not delete:
            return True
        if not remove(path):
            return False
    try:
        os.makedirs(path, mode)
        return True
    except Exception:
        logger.exception("Failed to mkdir: %s" % path)
        return False


def __create_destdir(destination):
    destdir = os.path.dirname(destination)
    if not os.path.isdir(destdir):
        if not mkdir(destdir):
            raise Exception('Failed to create "%s"' % destdir)


def __copyfile(source, destination):
    """Copy data and mode bits ("cp source destination").

    The destination may be a directory.

    Args:
        source (str): Source file (file to copy).
        destination (str): Destination file or directory (where to copy).

    Returns:
        bool: True if the operation is successful, False otherwise.
    """
    logger.info("copyfile: %s -> %s" % (source, destination))
    try:
        __create_destdir(destination)
        shutil.copy(source, destination)
        return True
    except Exception as e:
        logger.error(
            "copyfile: %s -> %s failed! Error: %s", source, destination, e
        )
        return False


def __copyfile2(source, destination):
    """Copy data and all stat info ("cp -p source destination").

    The destination may be a directory.

    Args:
        source (str): Source file (file to copy).
        destination (str): Destination file or directory (where to copy).

    Returns:
        bool: True if the operation is successful, False otherwise.
    """
    logger.info("copyfile2: %s -> %s" % (source, destination))
    try:
        __create_destdir(destination)
        shutil.copy2(source, destination)
        return True
    except Exception as e:
        logger.error(
            "copyfile2: %s -> %s failed! Error: %s", source, destination, e
        )
        return False


def __copytree(source, destination, symlinks=False):
    """Copy a directory tree recursively using copy2().

    The destination directory must not already exist.

    If the optional symlinks flag is true, symbolic links in the
    source tree result in symbolic links in the destination tree; if
    it is false, the contents of the files pointed to by symbolic
    links are copied.

    Args:
        source (str): Source directory (directory to copy).
        destination (str): Destination directory (where to copy).
        symlinks (bool): Follow symbolic links.

    Returns:
        bool: True if the operation is successful, False otherwise.
    """
    logger.info("copytree: %s -> %s" % (source, destination))
    try:
        __create_destdir(destination)
        shutil.copytree(source, destination, symlinks)
        return True
    except Exception as e:
        logger.exception(
            "copytree: %s -> %s failed! Error: %s", source, destination, e
        )
        return False


def copy(source, destination):
    """Copy file or directory.

    Args:
        source (str): Source file or directory
        destination (str): Destination file or directory (where to copy).

    Returns:
        bool: True if the operation is successful, False otherwise.
    """
    if os.path.isdir(source):
        return __copytree(source, destination)
    else:
        return __copyfile2(source, destination)


def gcopy(pattern, destination):
    """Copy all file found by glob.glob(pattern) to destination directory.

    Args:
        pattern (str): Glob pattern
        destination (str): Path to the destination directory.

    Returns:
        bool: True if the operation is successful, False otherwise.
    """
    for item in glob.glob(pattern):
        if not copy(item, destination):
            return False
    return True


def move(source, destination):
    """Move a file or directory (recursively) to another location.

    If the destination is on our current file system, then simply use
    rename. Otherwise, copy source to the destination and then remove
    source.

    Args:
        source (str): Source file or directory (file or directory to move).
        destination (str): Destination file or directory (where to move).

    Returns:
        bool: True if the operation is successful, False otherwise.
    """
    logger.info("Move: %s -> %s" % (source, destination))
    try:
        __create_destdir(destination)
        shutil.move(source, destination)
        return True
    except Exception:
        logger.exception("Failed to Move: %s -> %s" % (source, destination))
        return False


def gmove(pattern, destination):
    """Move all file found by glob.glob(pattern) to destination directory.

    Args:
        pattern (str): Glob pattern
        destination (str): Path to the destination directory.

    Returns:
        bool: True if the operation is successful, False otherwise.
    """
    for item in glob.glob(pattern):
        if not move(item, destination):
            return False
    return True


def __rmfile(path):
    """Delete a file.

    Args:
        path (str): Path to the file that needs to be deleted.

    Returns:
        bool: True if the operation is successful, False otherwise.
    """
    logger.info("rmfile: %s" % path)
    try:
        os.remove(path)
        return True
    except Exception as e:
        logger.error("rmfile: %s failed! Error: %s" % (path, e))
        return False


def __rmtree(path):
    """Recursively delete a directory tree.

    Args:
        path (str): Path to the directory that needs to be deleted.

    Returns:
        bool: True if the operation is successful, False otherwise.
    """
    logger.info("rmtree: %s" % path)
    try:
        shutil.rmtree(path)
        return True
    except Exception as e:
        logger.error("rmtree: %s failed! Error: %s" % (path, e))
        return False


def remove(path):
    """Delete a file or directory.

    Args:
        path (str): Path to the file or directory that needs to be deleted.

    Returns:
        bool: True if the operation is successful, False otherwise.
    """
    if os.path.isdir(path):
        return __rmtree(path)
    else:
        return __rmfile(path)


def gremove(pattern):
    """Remove all file found by glob.glob(pattern).

    Args:
        pattern (str): Pattern of files to remove
    Returns:
        bool: True if the operation is successful, False otherwise.
    """
    for item in glob.glob(pattern):
        if not remove(item):
            return False
    return True


def read(path, encoding="utf-8"):
    """Read the content of the file.

    Args:
        path (str): Path to the file
        encoding (str): File encoding. Default: utf-8

    Returns:
        str: File content or empty string if there was an error
    """
    try:
        with io.open(path, encoding=encoding) as f:
            return f.read()
    except Exception as e:
        logger.error("read: %s failed. Error: %s", path, e)
        return ""


def touch(path, content="", encoding="utf-8", overwrite=False):
    """Create a file at the given path if it does not already exists.

    Args:
        path (str): Path to the file.
        content (str): Optional content that will be written in the file.
        encoding (str): Encoding in which to write the content.
            Default: ``utf-8``
        overwrite (bool): Overwrite the file if exists.

    Returns:
        bool: True if the operation is successful, False otherwise.
    """
    path = os.path.abspath(path)
    if not overwrite and os.path.exists(path):
        logger.warning('touch: "%s" already exists', path)
        return False
    try:
        logger.info("touch: %s", path)
        with io.open(path, "wb") as f:
            if not isinstance(content, bytes):
                content = content.encode(encoding)
            f.write(content)
        return True
    except Exception as e:
        logger.error("touch: %s failed. Error: %s", path, e)
        return False



import threading


class SingletonMetaclass(type):
    """Singleton Metaclass.

    This metaclass is used for creating singletons.
    It changes the class __new__ method to maintain only one instance of the
    class, and tweaks the __init__ method to be executed only once (when the
    first instance of the class is created.

    Usage::

        >>> class MySingleton(object, metaclass=SingletonMetaclass):
        ...     '''Real singleton class.
        ...
        ...     You have to set the metaclass to SingletonMetaclass,
        ...     and define the __init__ function. Everything else will
        ...     be done by metaclass.
        ...     '''
        ...     def __init__(self, data):
        ...         print("Initializing")
        ...         self.data = data
        ...
        >>> # Only this actually happen
        >>> first = MySingleton("First initialization")
        Initializing
        >>> second = MySingleton("Second initialization") # This won't happen
        >>> first.data
        'First initialization'
        >>> second.data
        'First initialization'
    """

    def __init__(cls, *args, **kwargs):
        super(SingletonMetaclass, cls).__init__(*args, **kwargs)
        cls._instance = None
        cls._lock = threading.Lock()

    def __call__(cls, *args, **kwargs):
        with cls._lock:
            if cls._instance is None:
                cls._instance = super(SingletonMetaclass, cls).__call__(
                    *args, **kwargs
                )
            return cls._instance

    @property
    def instance(cls):
        with cls._lock:
            return cls._instance


class Singleton(metaclass=SingletonMetaclass):
    """Singleton class.

    Inherit from this class if you want to have a singleton class.

    Never use SingletonMetaclass!

    Usage::

        >>> class EmptySingleton(Singleton):
        ...     '''Singleton without __init__ method'''
        ...     pass
        ...
        >>> first = EmptySingleton()
        >>> second = EmptySingleton()
        >>> assert id(first) == id(second)
        >>>
        >>> class InitSingleton(Singleton):
        ...     '''Singleton with __init__ method'''
        ...     def __init__(self, data):
        ...         print("Initializing")
        ...         self.data = data
        ...
        >>> # Only this actually happen
        >>> first = InitSingleton("First initialization")
        Initializing
        >>> second = InitSingleton("Second initialization") # This won't happen
        >>> first.data
        'First initialization'
        >>> assert first.data == second.data
    """

    pass



from typing import Optional
from datetime import datetime, date

import pytz
import tzlocal


TIMESTAMP_FORMAT = "%Y-%m-%dT%H:%M:%S"
DATE_FORMAT = "%Y-%m-%d"
TIME_FORMAT = "%H:%M:%S"
HOURS = 60 * 60
MINUTES = 60


# Reuse so I don't need to import pytz too
utc = pytz.utc


def now():
    """Return an aware datetime.datetime now."""
    # timeit shows that datetime.now(tz=utc) is 24% slower
    return datetime.utcnow().replace(tzinfo=utc)


def is_aware(value):
    """Determine if a given datetime.datetime is aware.

    The concept is defined in Python's docs:
    https://docs.python.org/library/datetime.html#datetime.tzinfo

    Assuming value.tzinfo is either None or a proper datetime.tzinfo,
    value.utcoffset() implements the appropriate logic.
    """
    return value.utcoffset() is not None


def is_naive(value):
    """Determine if a given datetime.datetime is naive.

    The concept is defined in Python's docs:
    https://docs.python.org/library/datetime.html#datetime.tzinfo

    Assuming value.tzinfo is either None or a proper datetime.tzinfo,
    value.utcoffset() implements the appropriate logic.
    """
    return value.utcoffset() is None


def make_aware(value, timezone=None, is_dst=None):
    """Make a naive datetime.datetime in a given time zone aware."""
    if timezone is None:
        timezone = tzlocal.get_localzone()
    if hasattr(timezone, "localize"):
        # This method is available for pytz time zones.
        return timezone.localize(value, is_dst=is_dst)
    else:
        # Check that we won't overwrite the timezone of an aware datetime.
        if is_aware(value):
            raise ValueError(
                "make_aware expects a naive datetime, got %s" % value
            )
        # This may be wrong around DST changes!
        return value.replace(tzinfo=timezone)


def localtime(value=None, timezone=None):
    """
    Convert an aware datetime.datetime to local time.

    Only aware datetimes are allowed. When value is omitted, it defaults to
    now().

    Local time is defined by the current time zone, unless another time zone
    is specified.
    """
    if value is None:
        value = now()
    if timezone is None:
        timezone = tzlocal.get_localzone()
    # Emulate the behavior of astimezone() on Python < 3.6.
    if is_naive(value):
        raise ValueError("localtime() cannot be applied to a naive datetime")
    return value.astimezone(timezone)


def date_to_str(d: Optional[date]) -> Optional[str]:
    """Convert date to string."""
    if d is None:
        return None

    return d.strftime(DATE_FORMAT)


def date_to_datetime(d: Optional[date]) -> Optional[datetime]:
    """Convert a date to timezone aware datatetime."""
    if d is None:
        return None

    return make_aware(datetime.combine(d, datetime.min.time()))


def to_utc_str(dt: Optional[datetime]) -> Optional[str]:
    """Convert datetime to utc string."""
    if dt is None:
        return None

    if isinstance(dt, date):
        return dt.strftime(TIMESTAMP_FORMAT)

    if not is_aware(dt):
        dt = make_aware(dt, timezone=utc)
    return localtime(dt, timezone=utc).strftime(TIMESTAMP_FORMAT)


def to_localtime_str(dt: Optional[datetime]) -> Optional[str]:
    """Convert datetime to local time aware string."""
    if dt is None:
        return None

    if not is_aware(dt):
        dt = make_aware(dt, timezone=utc)
    return localtime(dt).strftime(TIME_FORMAT)


def from_utc_str(s: Optional[str]) -> Optional[datetime]:
    """Convert string to timezone aware datetime."""
    if s is None:
        return None
    return make_aware(datetime.strptime(s, TIMESTAMP_FORMAT), timezone=utc)


def humanize(duration: int) -> str:
    """Convert duration in seconds to human readable representation.

    Args:
        duration (int): Duration in seconds.
    """
    hours = duration // HOURS

    result = []
    if hours > 0:
        result.append(f"{hours}h")
        duration = duration - (HOURS * hours)

    minutes = duration // MINUTES
    if minutes > 0 or hours > 0:
        result.append(f"{minutes:02d}m")
        duration = duration - (MINUTES * minutes)

    result.append(f"{duration:02d}s")
    return " ".join(result)
